"""
This is the documentation for the python package [**pyseafuel**](https://pypi.org/project/pyseafuel/).
"""

# ──────────────────────────────────────────────────────────────────────────

__version__ = '0.1.0'

# ──────────────────────────────────────────────────────────────────────────

from pyseafuel import degasser
from pyseafuel import desalinator
from pyseafuel import dic
from pyseafuel import electrolyzer
from pyseafuel import reactor
from pyseafuel import system
